#!/usr/bin/env python3
#
# Copyright 2017,2021 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper APIs for interaction with Google Fonts.

Provides APIs to interact with font subsets, codepoints for font or subset.
"""

from __future__ import print_function
from __future__ import unicode_literals

import codecs
import collections
import contextlib
import errno
import os
import re
import sys
import glob
from pkg_resources import resource_filename

if __name__ == '__main__':
  # some of the imports here wouldn't work otherwise
  sys.path.append(
      os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import gftools.fonts_public_pb2 as fonts_pb2
from fontTools import ttLib
from gflanguages import LoadLanguages
from glyphsets.subsets import SUBSETS
from google.protobuf import text_format
from hyperglot import parse



# See https://www.microsoft.com/typography/otspec/name.htm.
NAME_COPYRIGHT = 0
NAME_FAMILY = 1
NAME_UNIQUEID = 3
NAME_FULLNAME = 4
NAME_PSNAME = 6


_PLATFORM_ID_MICROSOFT = 3
_PLATFORM_ENC_UNICODE_BMP = 1
_PLATFORM_ENC_UNICODE_UCS4 = 10
_PLATFORM_ENCS_UNICODE = (_PLATFORM_ENC_UNICODE_BMP, _PLATFORM_ENC_UNICODE_UCS4)

_FAMILY_WEIGHT_REGEX = r'([^/-]+)-(\w+)\.[ot]tf$'

# Matches 4 or 5 hexadecimal digits that are uppercase at the beginning of the
# test string. The match is stored in group 0, e.g:
# >>> _NAMELIST_CODEPOINT_REGEX.match('1234X').groups()[0]
# '1234'
# >>> _NAMELIST_CODEPOINT_REGEX.match('1234A').groups()[0]
# '1234A'
_NAMELIST_CODEPOINT_REGEX = re.compile('^([A-F0-9]{4,5})')

# The canonical [to Google Fonts] name comes before any aliases
_KNOWN_WEIGHTS = collections.OrderedDict([
    ('Thin', 100),
    ('Hairline', 100),
    ('ExtraLight', 200),
    ('Light', 300),
    ('Regular', 400),
    ('', 400),  # Family-Italic resolves to this
    ('Medium', 500),
    ('SemiBold', 600),
    ('Bold', 700),
    ('ExtraBold', 800),
    ('Black', 900)
])

_VALID_STYLES = {'normal', 'italic'}

# (Mask, Name) pairs.
# See https://www.microsoft.com/typography/otspec/os2.htm#fss.
_FS_SELECTION_BITS = tuple(
    (1 << i, n)
    for i, n in enumerate(('ITALIC', 'UNDERSCORE', 'NEGATIVE', 'OUTLINED',
                           'STRIKEOUT', 'BOLD', 'REGULAR', 'USE_TYPO_METRICS',
                           'WWS', 'OBLIQUE')))


# license_dir => license name mappings
_KNOWN_LICENSE_DIRS = {
    'apache': 'APACHE2',
    'ofl': 'OFL',
    'ufl': 'UFL',
}


FileFamilyStyleWeightTuple = collections.namedtuple(
    'FileFamilyStyleWeightTuple', ['file', 'family', 'style', 'weight'])


class Error(Exception):
  """Base for Google Fonts errors."""


class ParseError(Error):
  """Exception used when parse failed."""


def UnicodeCmapTables(font):
  """Find unicode cmap tables in font.

  Args:
    font: A TTFont.
  Yields:
    cmap tables that contain unicode mappings
  """
  for table in font['cmap'].tables:
    if (table.platformID == _PLATFORM_ID_MICROSOFT and
        table.platEncID in _PLATFORM_ENCS_UNICODE):
      yield table


def UniqueSort(*args):
  """Returns a sorted list of the unique items from provided iterable(s).

  Args:
    *args: Iterables whose items will be merged, sorted and de-duplicated.
  Returns:
    A list.
  """
  s = set()
  for arg in args:
    s.update(arg)
  return sorted(s)


def RegularWeight(metadata):
  """Finds the filename of the regular (normal/400) font file.

  Args:
    metadata: The metadata to search for the regular file data.
  Returns:
    The name of the regular file, usually Family-Regular.ttf.
  Raises:
    OSError: If regular file could not be found. errno.ENOENT.
  """
  for f in metadata.fonts:
    if f.weight == 400 and f.style == 'normal':
      return os.path.splitext(f.filename)[0] + '.ttf'

  name = '??'
  if metadata.HasField('name'):
    name = metadata.name
  raise OSError(errno.ENOENT, 'unable to find regular weight in %s' % name)


def ListSubsets():
  """Returns a list of all subset names, in lowercase."""
  return SUBSETS


def Metadata(file_or_dir):
  """Returns fonts_metadata.proto object for a metadata file.

  If file_or_dir is a file named METADATA.pb, load it. If file_or_dir is a
  directory, load the METADATA.pb file in that directory.

  Args:
    file_or_dir: A file or directory.
  Returns:
    Python object loaded from METADATA.pb content.
  Raises:
    ValueError: if file_or_dir isn't a METADATA.pb file or dir containing one.
  """
  if (os.path.isfile(file_or_dir) and
      os.path.basename(file_or_dir) == 'METADATA.pb'):
    metadata_file = file_or_dir
  elif os.path.isdir(file_or_dir):
    metadata_file = os.path.join(file_or_dir, 'METADATA.pb')
    if not os.path.isfile(metadata_file):
      raise ValueError('No METADATA.pb in %s' % file_or_dir)
  else:
    raise ValueError(
        '%s is neither METADATA.pb file or a directory' % file_or_dir)

  msg = fonts_pb2.FamilyProto()
  with codecs.open(metadata_file, encoding='utf-8') as f:
    text_format.Merge(f.read(), msg)

  return msg


def FamilyName(fontname):
  """Attempts to build family name from font name.

  For example, HPSimplifiedSans => HP Simplified Sans.

  Args:
    fontname: The name of a font.
  Returns:
    The name of the family that should be in this font.
  """
  # SomethingUpper => Something Upper
  fontname = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', fontname)
  # Font3 => Font 3
  fontname = re.sub('([a-z])([0-9]+)', r'\1 \2', fontname)
  # lookHere => look Here
  return re.sub('([a-z0-9])([A-Z])', r'\1 \2', fontname)


def Weight(stylename):
  """Derive weight from a stylename.

  Args:
    stylename: string, e.g. Bold, Regular, or ExtraLightItalic.
  Returns:
    weight: integer
  """
  if stylename.endswith('Italic'):
    return _KNOWN_WEIGHTS[stylename[:-6]]
  return _KNOWN_WEIGHTS[stylename]


def VFWeight(font):
  """Return a variable fonts weight. Return 400 if 400 is within the wght
  axis range else return the value closest to 400

  Args:
    font: TTFont
  Returns:
    weight: integer
  """
  wght_axis = None
  for axis in font['fvar'].axes:
    if axis.axisTag == "wght":
      wght_axis = axis
      break
  value = 400
  if wght_axis:
    if wght_axis.minValue >= 400:
      value = wght_axis.minValue
    if wght_axis.maxValue <= 400:
      value = wght_axis.maxValue
  # TODO (MF) check with GF Eng if we should just assume it's safe to return
  # 400 if a wght axis doesn't exist.
  return int(value)


def Style(stylename):
  return 'italic' if "Italic" in stylename else "normal"


def FamilyStyleWeight(path):
    filename = os.path.basename(path)
    if "[" in filename and "]" in filename:
        return VFFamilyStyleWeight(path)
    return FileFamilyStyleWeight(path)


def FileFamilyStyleWeight(path):
  """Extracts family, style, and weight from Google Fonts standard filename.

  Args:
    path: Font path, eg ./fonts/ofl/lobster/Lobster-Regular.ttf.
  Returns:
    FileFamilyStyleWeightTuple for file.
  Raises:
    ParseError: if file can't be parsed.
  """
  m = re.search(_FAMILY_WEIGHT_REGEX, path)
  if not m:
    raise ParseError('Could not parse %s' % path)
  style = Style(m.group(2))
  weight = Weight(m.group(2))
  return FileFamilyStyleWeightTuple(path, FamilyName(m.group(1)), style,
                                    weight)


def VFFamilyStyleWeight(path):
  """Extract family, style and weight from a variable font's name table.

  Args:
      path: Font path, eg ./fonts/ofl/lobster/Lobster[wght].ttf.
  Returns:
    FileFamilyStyleWeightTuple for file.
  """
  with ttLib.TTFont(path) as font:
    typoFamilyName = font['name'].getName(16, 3, 1, 1033)
    familyName = font['name'].getName(1, 3, 1, 1033)
    family = typoFamilyName.toUnicode() if typoFamilyName else \
             familyName.toUnicode()

    typoStyleName = font['name'].getName(17, 3, 1, 1033)
    styleName = font['name'].getName(2, 3, 1, 1033)
    style = typoStyleName.toUnicode() if typoStyleName else \
            styleName.toUnicode()
    style = "italic" if "Italic" in style.replace(" ", "") else "normal"
    # For each font in a variable font family, we do not want to return
    # the style's weight. We want to return 400 if 400 is within the
    # the wght axis range. If it isn't, we want the value closest to 400.
    weight = VFWeight(font)
    return FileFamilyStyleWeightTuple(path, family, style, weight)


def ExtractNames(font, name_id):
  return [
      n.toUnicode()
      for n in font['name'].names
      if n.nameID == name_id
  ]


def ExtractName(font_or_file, name_id, default):
  """Extracts a name table field (first value if many) from a font.

  Args:
    font_or_file: path to a font file or a TTFont.
    name_id: the ID of the name desired. Use NAME_* constant.
    default: result if no value is present.
  Returns:
    The value of the first entry for name_id or default if there isn't one.
  """
  value = default
  names = []
  if isinstance(font_or_file, ttLib.TTFont):
    names = ExtractNames(font_or_file, name_id)
  else:
    with contextlib.closing(ttLib.TTFont(font_or_file)) as font:
      names = ExtractNames(font, name_id)

  if names:
    value = names[0]

  return value


def NamePartsForStyleWeight(astyle, aweight):
  """Gives back the parts that go into the name for this style/weight.

  Args:
    astyle: The style name, eg "normal" or "italic"
    aweight: The font weight
  Returns:
    Tuple of parts that go into the name, typically the name for the weight and
    the name for the style, if any ("normal" typically doesn't factor into
    names).
  Raises:
    ValueError: If the astyle or aweight isn't a supported value.
  """
  astyle = astyle.lower()
  if astyle not in _VALID_STYLES:
    raise ValueError('unsupported style %s' % astyle)

  correct_style = None
  if astyle == 'italic':
    correct_style = 'Italic'

  correct_name = None
  for name, weight in _KNOWN_WEIGHTS.items():
    if weight == aweight:
      correct_name = name
      break

  if not correct_name:
    raise ValueError('unsupported weight: %d' % aweight)

  return tuple([n for n in [correct_name, correct_style] if n])


def _RemoveAll(alist, value):
  while value in alist:
    alist.remove(value)


def FilenameFor(family, style, weight, ext=''):
  family = family.replace(' ', '')
  style_weight = list(NamePartsForStyleWeight(style, weight))
  if 'Italic' in style_weight:
    _RemoveAll(style_weight, 'Regular')

  style_weight = ''.join(style_weight)
  return '%s-%s%s' % (family, style_weight, ext)


def FullnameFor(family, style, weight):
  name_parts = [family]
  name_parts.extend(list(NamePartsForStyleWeight(style, weight)))
  _RemoveAll(name_parts, 'Regular')
  return ' '.join(name_parts)


def FontDirs(path):
  """Finds all the font directories (based on METADATA.pb) under path.

  Args:
    path: A path to search under.
  Yields:
    Directories under path that have a METADATA.pb.
  """
  for dir_name, _, _ in os.walk(path):
    if os.path.isfile(os.path.join(dir_name, 'METADATA.pb')):
      yield dir_name


def FsSelectionMask(flag):
  """Get the mask for a given named bit in fsSelection.

  Args:
    flag: Name of the flag per otspec, eg ITALIC, BOLD, etc.
  Returns:
    Bitmask for that flag.
  Raises:
    ValueError: if flag isn't the name of any fsSelection bit.
  """
  for (mask, name) in _FS_SELECTION_BITS:
    if name == flag:
      return mask
  raise ValueError('No mask for %s' % flag)


def FsSelectionFlags(fs_selection):
  """Get the named flags enabled in a given fsSelection.

  Args:
    fs_selection: An fsSelection value.
  Returns:
    List of names of flags enabled in fs_selection.
  """
  names = []
  for (mask, name) in _FS_SELECTION_BITS:
    if fs_selection & mask:
      names.append(name)
  return names


def _EntryForEndOfPath(path, answer_map):
  segments = [s.lower() for s in path.split(os.sep)]
  answers = [answer_map[s] for s in segments if s in answer_map]
  if len(answers) != 1:
    raise ValueError('Found %d possible matches: %s' % (len(answers), answers))
  return answers[0]


def LicenseFromPath(path):
  """Try to figure out the license for a given path.

  Splits path and looks for known license dirs in segments.

  Args:
    path: A filesystem path, hopefully including a license dir.
  Returns:
    The name of the license, eg OFL, UFL, etc.
  Raises:
    ValueError: if 0 or >1 licenses match path.
  """
  return _EntryForEndOfPath(path, _KNOWN_LICENSE_DIRS)


# Note:      This function uses hyperglot, which is licensed GPLv3
# See also:  https://github.com/googlefonts/gftools/issues/498
def SupportedLanguages(ttFont, languages=LoadLanguages()):
  """Get languages supported by given ttFont.

  Languages are pulled from the given set. Based on whether exemplar character
  sets are present in the given font.

  Logic based on Hyperglot: https://github.com/rosettatype/hyperglot/blob/3172061ca05a62c0ff330eb802a17d4fad8b1a4d/lib/hyperglot/language.py#L273-L301
  """
  chars = [chr(c) for c in ttFont["cmap"].getBestCmap()]
  supported = []
  for lang in languages.values():
    if not lang.HasField('exemplar_chars') or not lang.exemplar_chars.HasField('base'):
      continue
    base = parse.parse_chars(lang.exemplar_chars.base,
                             decompose=False,
                             retainDecomposed=False)
    if set(base).issubset(chars):
      supported.append(lang)
  return supported


def GetExemplarFont(family):
  assert len(family.fonts) > 0, 'Unable to select exemplar in family with no fonts: ' + family.name
  for font in family.fonts:
    if font.style == 'normal' and font.weight == 400:
      # Prefer default style (Regular, not Italic)
      return font
  return family.fonts[0]


def LanguageComments(languages):
  """Generate a mapping for METADATA.pb language field comments.
  Every language field in a METADATA.pb has a comment which is the language's name e.g
    languages: "xh_Latn"  # Xhosa

  Args:
    languages: a dict with keys for lang tags and values for fonts_public_pb2.LanguageProto objects e.g
      languages={"kr_Arab": <class 'fonts_public_pb2.LanguageProto'>}
  Returns:
    A dict with keys for the language field entry and values for the comment.
        {
        'languages: "kr_Arab"': 'Kanuri',
        'languages: "fi_Latn"': 'Finnish'
      }
  """
  line_to_lang_name = {}
  for language in languages.values():
    line = f'languages: "{language.id}"'
    line_to_lang_name[line] = language.name
  return line_to_lang_name


def ReadProto(proto, path):
  with open(path, 'r', encoding='utf-8') as f:
    proto = text_format.Parse(f.read(), proto)
    return proto


def WriteProto(proto, path, comments = None):
  with open(path, 'w', newline='') as f:
    textproto = text_format.MessageToString(proto, as_utf8=True)
    if comments is not None:
      lines = [s if s not in comments else s + '  # ' + comments[s] for s in textproto.split('\n')]
      textproto = '\n'.join(lines)
    f.write(textproto)



#!/usr/bin/env python3
"""
Fix designspace instances so they are gf spec complaint:
https://github.com/googlefonts/gf-docs/tree/main/Spec#fvar-instances

Usage:

gftools ufo-fix-instances family.designspace
"""
from fontTools.designspaceLib import DesignSpaceDocument, InstanceDescriptor
from copy import deepcopy
from gftools.fix import WEIGHT_VALUES
from axisregistry import AxisRegistry
import sys

axis_reg = AxisRegistry()

def build_instances(ds):
    """Generate Designspace instances which are gf spec complaint"""
    instances = []
    axes = ds.axes
    wght_axis = next((a for a in axes if a.tag == "wght"), None)
    wght_vals = list(range(int(wght_axis.minimum), int(wght_axis.maximum + 100), 100))
    italic_axis = next((a for a in axes if a.tag in ["slnt", "ital"]), None)

    dflt_coords = {}
    for axis in axes:
        if axis.tag not in axis_reg or (
            axis_reg[axis.tag].default_value < axis.minimum
            or axis_reg[axis.tag].default_value > axis.maximum
        ):
            dflt_coords[axis.tag] = axis.map_forward(axis.default)
        else:
            dflt_coords[axis.tag] = axis.map_forward(axis_reg[axis.tag].default_value)

    for wght in wght_vals:
        inst = InstanceDescriptor()
        coords = deepcopy(dflt_coords)
        coords["wght"] = wght_axis.map_forward(wght)
        inst.location = coords
        inst.styleName = WEIGHT_VALUES[wght]
        instances.append(inst)

    if italic_axis:
        italics = deepcopy(instances)
        for inst in italics:
            inst.styleName = f"{inst.styleName} Italic".replace(
                "Regular Italic", "Italic"
            )
            inst.location[italic_axis.tag] = italic_axis.map_forward(
                italic_axis.minimum
            )
        instances += italics
    return instances


def main():
    if len(sys.argv) != 2:
        print(__doc__)
        return
    ds_path = sys.argv[1]
    ds = DesignSpaceDocument()
    ds.read(ds_path)
    instances = build_instances(ds)
    ds.instances = instances
    ds.write(ds_path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
gftools lang

Adds sample text for a given language using the specified UDHR translation.

Usage:

# Standard usage. Output lang metadata to a dir. Does not overwrite existing data.
gftools lang-sample-text -l ./languages/en.textproto ./udhr_translations/en.xml

"""

from absl import app
from absl import flags
from gflanguages import (LoadLanguages,
                         LoadRegions)
from gftools import fonts_public_pb2
from gftools.util.udhr import Udhr
from google.protobuf import text_format
from lxml import etree
import csv
import glob
import os
import re
import yaml

FLAGS = flags.FLAGS
flags.DEFINE_string('lang', None, 'Path to lang metadata package', short_name='l')
flags.DEFINE_string('udhrs', None, 'Path to UDHR translations (XML)', short_name='u')
flags.DEFINE_string('samples', None, 'Path to per-family samples from noto-data-dev repo', short_name='s')

def _ReadProto(proto, path):
  with open(path, 'r', encoding='utf-8') as f:
    proto = text_format.Parse(f.read(), proto)
    return proto


def _WriteProto(proto, path, comments = None):
  with open(path, 'w', newline='') as f:
    textproto = text_format.MessageToString(proto, as_utf8=True)
    if comments is not None:
      lines = [s if s not in comments else s + '  # ' + comments[s] for s in textproto.split('\n')]
      textproto = '\n'.join(lines)
    f.write(textproto)


def _GetLanguageForUdhr(languages, udhr):
  for l in languages.values():
    if (l.language == udhr.iso639_3 and l.script == udhr.iso15924) or \
        l.id == udhr.bcp47:
      return l

  language = fonts_public_pb2.LanguageProto()
  language.id = udhr.bcp47
  language.language = udhr.bcp47.split('_')[0]
  language.script = udhr.iso15924
  language.name = udhr.name.replace(' (', ', ').replace(')', '')
  return language


def _ReplaceInSampleText(languages):
  for l in languages.values():
    if l.script == 'Latn' or not l.HasField('sample_text'):
      continue
    if '-' in l.sample_text.masthead_full:
      l.sample_text.masthead_full = l.sample_text.masthead_full.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.masthead_partial:
      l.sample_text.masthead_partial = l.sample_text.masthead_partial.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.styles:
      l.sample_text.styles = l.sample_text.styles.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.tester:
      l.sample_text.tester = l.sample_text.tester.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.poster_sm:
      l.sample_text.poster_sm = l.sample_text.poster_sm.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.poster_md:
      l.sample_text.poster_md = l.sample_text.poster_md.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.poster_lg:
      l.sample_text.poster_lg = l.sample_text.poster_lg.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.specimen_48:
      l.sample_text.specimen_48 = l.sample_text.specimen_48.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.specimen_36:
      l.sample_text.specimen_36 = l.sample_text.specimen_36.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.specimen_32:
      l.sample_text.specimen_32 = l.sample_text.specimen_32.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.specimen_21:
      l.sample_text.specimen_21 = l.sample_text.specimen_21.replace(' - ', ' ').replace('-', '').strip()
    if '-' in l.sample_text.specimen_16:
      l.sample_text.specimen_16 = l.sample_text.specimen_16.replace(' - ', ' ').replace('-', '').strip()

    _WriteProto(l, os.path.join(FLAGS.lang, 'languages', l.id + '.textproto'))


def main(argv):
  languages = LoadLanguages(base_dir=FLAGS.lang)
  regions = LoadRegions(base_dir=FLAGS.lang)

  if FLAGS.samples:
    assert len(argv) > 1, 'No METADATA.pb files specified'
    line_to_lang_name = {}
    for l in languages:
      line = 'languages: "{code}"'.format(code=languages[l].id)
      line_to_lang_name[line] = languages[l].name
    samples = {}
    for sample_filename in os.listdir(FLAGS.samples):
      key = os.path.splitext(os.path.basename(sample_filename))[0]
      samples[key] = os.path.join(FLAGS.samples, sample_filename)
    for path in argv[1:]:
      family = _ReadProto(fonts_public_pb2.FamilyProto(), path)
      if True:#len(family.languages) == 0 or family.name == 'Noto Sans Tamil Supplement':
        key = family.name.replace(' ', '')
        if key not in samples:
          print('Family not found in samples: ' + family.name)
          continue
        with open(samples[key], 'r') as f:
          sample_data = yaml.safe_load(f)
          sample_text = fonts_public_pb2.SampleTextProto()
          sample_text.masthead_full = sample_data['masthead_full']
          sample_text.masthead_partial = sample_data['masthead_partial']
          sample_text.styles = sample_data['styles']
          sample_text.tester = sample_data['tester']
          sample_text.poster_sm = sample_data['poster_sm']
          sample_text.poster_md = sample_data['poster_md']
          sample_text.poster_lg = sample_data['poster_lg']
          family.sample_text.MergeFrom(sample_text)
          _WriteProto(family, path, comments=line_to_lang_name)

  if not FLAGS.udhrs:
    return

  if FLAGS.udhrs.endswith('.yaml'):
    with open(FLAGS.udhrs, 'r') as f:
      data = yaml.safe_load(f)
      for translation, meta in data.items():
        if 'lang_full' not in meta or meta['lang_full'] not in ['ccp-Beng-IN', 'lad-Hebr-IL']:
          continue
        language = meta['lang']
        if language.startswith('und-'):
          continue
        script = re.search(r'.*-(.*)-.*', meta['lang_full']).group(1) if 'script' not in meta else meta['script']
        key = language + '_' + script
        iso639_3 = meta['lang_639_3']
        iso15924 = script
        name = meta['name_lang'] if 'name_udhr' not in meta else meta['name_udhr']
        udhr = Udhr(
          key=key,
          iso639_3=iso639_3,
          iso15924=iso15924,
          bcp47=key,
          direction=None,
          ohchr=None,
          stage=4,
          loc=None,
          name=name
        )
        udhr.LoadArticleOne(translation)

        language = _GetLanguageForUdhr(languages, udhr)
        if not language.HasField('sample_text'):
          language.sample_text.MergeFrom(udhr.GetSampleTexts())
        if 'name_autonym' in meta and not language.HasField('autonym'):
          language.autonym = meta['name_autonym'].strip()
        _WriteProto(language, os.path.join(FLAGS.lang, 'languages', language.id + '.textproto'))

  elif FLAGS.udhrs.endswith('.csv'):
    with open(FLAGS.udhrs, newline='') as csvfile:
      reader = csv.reader(csvfile, delimiter=',', quotechar='"')
      head = next(reader)
      index_id = head.index('id')
      index_name = head.index('language')
      index_historical = head.index('historical')
      index_sample = head.index('SAMPLE')
      for row in reader:
        id = row[index_id]
        if id in languages:
          language = languages[row[index_id]]
        else:
          language = fonts_public_pb2.LanguageProto()
          language.id = id
          language.language, language.script = id.split('_')
          language.name = row[index_name]
        historical = row[index_historical] == 'X'
        if language.historical != historical:
          if historical:
            language.historical = True
          else:
            language.ClearField('historical')
        sample = row[index_sample]
        if sample and not sample.startswith('http'):
          udhr = Udhr(
            key=id,
            iso639_3=language.language,
            iso15924=language.script,
            bcp47=id,
            direction=None,
            ohchr=None,
            stage=4,
            loc=None,
            name=None
          )
          udhr.LoadArticleOne(sample)
          if not language.HasField('sample_text'):
            language.sample_text.MergeFrom(udhr.GetSampleTexts())
        _WriteProto(language, os.path.join(FLAGS.lang, 'languages', language.id + '.textproto'))

  elif os.path.isdir(FLAGS.udhrs):
    for udhr_path in glob.glob(os.path.join(FLAGS.udhrs, '*')):
      if udhr_path.endswith('index.xml') or os.path.basename(udhr_path).startswith('status'):
        continue
      udhr_data = etree.parse(udhr_path)
      head = udhr_data.getroot()
      for name, value in head.attrib.items():
        if re.search(r'\{.*\}lang', name):
          bcp47 = value.replace('-', '_')
      udhr = Udhr(
          key=head.get('key'),
          iso639_3=head.get('iso639-3'),
          iso15924=head.get('iso15924'),
          bcp47=bcp47,
          direction=head.get('dir'),
          ohchr=None,
          stage=4,
          loc=None,
          name=head.get('n'))
      udhr.Parse(udhr_data)

      language = _GetLanguageForUdhr(languages, udhr)
      if language.id in languages or language.HasField('sample_text'):
        continue
      language.sample_text.MergeFrom(udhr.GetSampleTexts())
      _WriteProto(language, os.path.join(FLAGS.lang, 'languages', language.id + '.textproto'))

  else:
    raise Exception('Unsupported input type for --udhrs: ' + FLAGS.udhrs)


if __name__ == '__main__':
  app.run(main)



import pytest
import os
from glob import glob
from gftools.stat import *
from fontTools.ttLib import TTFont
import yaml


TEST_DATA = os.path.join("data", "test")


@pytest.fixture
def var_font():
    """VF family consisting of a single font with two axes, wdth, wght"""
    return TTFont(os.path.join(TEST_DATA, "Inconsolata[wdth,wght].ttf"))


@pytest.fixture
def var_fonts():
    """VF family consisting of two fonts, Roman, and Italic. Both have a
    weight axis"""
    paths = [
        os.path.join(TEST_DATA, "Raleway[wght].ttf"),
        os.path.join(TEST_DATA, "Raleway-Italic[wght].ttf")
    ]
    return [TTFont(p) for p in paths]


@pytest.fixture
def var_fonts2():
    """VF family consisting of four fonts, Roman, Italic, Condensed Roman,
    Condensed Italic. All only have a wght axis"""
    paths = [
        os.path.join(TEST_DATA, "cabin_split", "Cabin[wght].ttf"),
        os.path.join(TEST_DATA, "cabin_split", "Cabin-Italic[wght].ttf"),
        os.path.join(TEST_DATA, "cabin_split", "CabinCondensed[wght].ttf"),
        os.path.join(TEST_DATA, "cabin_split", "CabinCondensed-Italic[wght].ttf")
    ]
    return [TTFont(p) for p in paths]


@pytest.fixture
def var_fonts3():
    """VF family consisting of two fonts, Roman and Italic. Both have wdth and wght
    axies."""
    paths = [
        os.path.join(TEST_DATA, "cabin_multi", "Cabin[wdth,wght].ttf"),
        os.path.join(TEST_DATA, "cabin_multi", "Cabin-Italic[wdth,wght].ttf")
    ]
    return [TTFont(p) for p in paths]


def _get_axis_value(font, axis, name, value):
    nametable = font["name"]
    stat = font["STAT"].table
    axis_indexes = {i: a.AxisTag for i,a in enumerate(stat.DesignAxisRecord.Axis)}
    axis_values = stat.AxisValueArray.AxisValue 
    for axis_value in axis_values:
        axis_tag = axis_indexes[axis_value.AxisIndex]
        if axis_tag != axis:
            continue
        nameID = axis_value.ValueNameID
        name_string = nametable.getName(nameID, 3, 1, 0x409).toUnicode()
        if name_string != name:
            continue
        if axis_value.Value != value:
            continue
        return axis_value
    return None


def test_gen_stat(var_font):
    del var_font["STAT"]
    gen_stat_tables([var_font], axis_order=["wdth", "wght"])
    stat = var_font["STAT"].table
    axes = {i: a.AxisTag for i,a in enumerate(stat.DesignAxisRecord.Axis)}

    axis_values = var_font['STAT'].table.AxisValueArray.AxisValue 
    # Check both wght and wdth axes exist
    axes_in_axis_values = set(axes[a.AxisIndex] for a in axis_values)
    assert axes_in_axis_values == {"wght", "wdth"}

    # Check wght axis values
    wght_axis_values = [v for v in axis_values if axes[v.AxisIndex] == "wght"]
    # Inconsolata has a min fvar wght of 200 and a max of 900.
    for axis_value, desired_weight in zip(wght_axis_values, range(200, 1000, 100)):
        assert axis_value.Value == desired_weight

    # Check wdth axis values
    wdth_axis_values = [v for v in axis_values if axes[v.AxisIndex] == "wdth"]
    # Inconsolata has a min fvar wdth of 50 and a max of 200.
    expected_wdths = [50, 62.5, 75, 87.5, 100, 112.5, 125, 150, 200]
    for axis_value, width in zip(wdth_axis_values, expected_wdths):
        assert axis_value.Value == width

    # Check Regular is linked to Bold
    reg_axis_value = _get_axis_value(var_font, "wght", "Regular", 400)
    assert reg_axis_value.LinkedValue == 700

    # Check Regular is elided
    assert reg_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME == ELIDABLE_AXIS_VALUE_NAME

    # Check Normal width is elided
    normal_axis_value = _get_axis_value(var_font, "wdth", "Normal", 100)
    assert normal_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME == ELIDABLE_AXIS_VALUE_NAME


def test_gen_stat_linked_values(var_font):
    del var_font["STAT"]
    gen_stat_tables([var_font], axis_order=["wdth", "wght"])
    stat = var_font["STAT"].table

    reg_axis_value = _get_axis_value(var_font, "wght", "Regular", 400)
    assert reg_axis_value.LinkedValue == 700


def test_gen_stat_linked_values_2(var_fonts2):
    gen_stat_tables(var_fonts2, axis_order=["wdth", "wght", "ital"])
    for font in var_fonts2:
        stat = font["STAT"].table
        reg_axis_value = _get_axis_value(font, "wght", "Regular", 400)
        assert reg_axis_value.LinkedValue == 700


def test_gen_stat_dflt_elided_values(var_fonts3):
    gen_stat_tables(var_fonts3, axis_order=["wdth", "wght", "ital"])
    for font in var_fonts3:
        stat = font["STAT"].table
        # Check regular axis value is elided
        reg_axis_value = _get_axis_value(font, "wght", "Regular", 400)
        assert reg_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME == ELIDABLE_AXIS_VALUE_NAME

        # Check normal (wdth) axis value is elided
        normal_axis_value = _get_axis_value(font, "wdth", "Normal", 100)
        assert normal_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME == ELIDABLE_AXIS_VALUE_NAME


def test_gen_stat_user_elided_values(var_fonts3):
    gen_stat_tables(
        var_fonts3,
        axis_order=["wdth", "wght", "ital"],
        elided_axis_values={"wght": [700], "wdth": [75]}
    )

    for font in var_fonts3:
        stat = font["STAT"].table
        # First check that the dflt axis values are not elided!
        reg_axis_value = _get_axis_value(font, "wght", "Regular", 400)
        assert reg_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME != ELIDABLE_AXIS_VALUE_NAME

        normal_axis_value = _get_axis_value(font, "wdth", "Normal", 100)
        assert normal_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME != ELIDABLE_AXIS_VALUE_NAME

        # now check the user specified elided values are elided
        bold_axis_value = _get_axis_value(font, "wght", "Bold", 700)
        assert bold_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME == ELIDABLE_AXIS_VALUE_NAME

        condensed_axis_value = _get_axis_value(font, "wdth", "Condensed", 75)
        assert condensed_axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME == ELIDABLE_AXIS_VALUE_NAME


def test_gen_stat_roman_and_italic_family(var_fonts):
    for var_font in var_fonts:
        del var_font["STAT"]
    gen_stat_tables(var_fonts, axis_order=["wght", "ital"])
    roman, italic = var_fonts

    roman_axis_val = _get_axis_value(roman, "ital", "Roman", 0.0)
    assert roman_axis_val != None
    assert roman_axis_val.LinkedValue == 1.0

    italic_axis_val = _get_axis_value(italic, "ital", "Italic", 1.0)
    assert italic_axis_val != None


def test_gen_stat_roman_and_italic_and_condensed_family(var_fonts2):
    gen_stat_tables(var_fonts2, axis_order=["wdth", "wght", "ital"])
    roman, italic, condensed_roman, condensed_italic = var_fonts2

    roman_ital_axis_val = _get_axis_value(roman, "ital", "Roman", 0.0)
    roman_wdth_axis_val = _get_axis_value(roman, "wdth", "Normal", 100.0)
    assert roman_ital_axis_val != None
    assert roman_wdth_axis_val != None

    italic_ital_axis_val = _get_axis_value(italic, "ital", "Italic", 1.0)
    italic_wdth_axis_val = _get_axis_value(italic, "wdth", "Normal", 100.0)
    assert italic_ital_axis_val != None
    assert italic_wdth_axis_val != None

    condensed_ital_axis_val = _get_axis_value(condensed_roman, "ital", "Roman", 0.0)
    condensed_wdth_axis_val = _get_axis_value(condensed_roman, "wdth", "Condensed", 75.0)
    assert condensed_ital_axis_val != None
    assert condensed_wdth_axis_val != None

    condensed_italic_ital_axis_val = _get_axis_value(condensed_italic, "ital", "Italic", 1.0)
    condensed_italic_wdth_axis_val = _get_axis_value(condensed_italic, "wdth", "Condensed", 75.0)
    assert condensed_italic_ital_axis_val != None
    assert condensed_italic_wdth_axis_val != None


def test_gen_stat_family_with_uneven_axes(var_fonts3):
    from fontTools.varLib.instancer import instantiateVariableFont
    roman, italic = var_fonts3
    # Drop the width axis from the roman font
    roman = instantiateVariableFont(roman, {"wdth": None})
    # We cannot add STAT tables to these families since the Google Fonts API
    # doesn't support them
    with pytest.raises(ValueError, match="fvar axes are not consistent across the family"):
        gen_stat_tables([roman, italic], axis_order=["wdth", "wght", "ital"])


def _check_ps_instance_names(ttfont, desired_names):
    nametable = ttfont['name']
    instances = ttfont['fvar'].instances
    for instance, desired_name in zip(instances, desired_names):
        ps_id = instance.postscriptNameID
        name = nametable.getName(ps_id, 3, 1, 0x409).toUnicode()
        assert name == desired_name


def test_gen_stat_update_fvar_instances_1(var_fonts):
    gen_stat_tables(var_fonts, axis_order=["wght", "ital"])
    roman, italic = var_fonts

    desired_roman_ps_names = [
        "RalewayRoman-Thin",
        "RalewayRoman-ExtraLight",
        "RalewayRoman-Light",
        "RalewayRoman-Regular",
        "RalewayRoman-Medium",
        "RalewayRoman-SemiBold",
        "RalewayRoman-Bold",
        "RalewayRoman-ExtraBold",
        "RalewayRoman-Black",
    ]
    _check_ps_instance_names(roman, desired_roman_ps_names)

    desired_italic_ps_names = [
        "RalewayItalic-Thin",
        "RalewayItalic-ExtraLight",
        "RalewayItalic-Light",
        "RalewayItalic-Regular",
        "RalewayItalic-Medium",
        "RalewayItalic-SemiBold",
        "RalewayItalic-Bold",
        "RalewayItalic-ExtraBold",
        "RalewayItalic-Black",
    ]
    _check_ps_instance_names(italic, desired_italic_ps_names)


def test_gen_stat_update_fvar_instances_2(var_fonts2):
    gen_stat_tables(var_fonts2, axis_order=["wdth", "wght", "ital"])
    roman, italic, condensed_roman, condensed_italic = var_fonts2

    desired_roman_ps_names = [
        "CabinNormalRoman-Regular",
        "CabinNormalRoman-Medium",
        "CabinNormalRoman-SemiBold",
        "CabinNormalRoman-Bold",
    ]
    _check_ps_instance_names(roman, desired_roman_ps_names)

    desired_italic_ps_names = [
        "CabinNormalItalic-Regular",
        "CabinNormalItalic-Medium",
        "CabinNormalItalic-SemiBold",
        "CabinNormalItalic-Bold",
    ]
    _check_ps_instance_names(italic, desired_italic_ps_names)

    desired_condensed_roman_ps_names = [
        "CabinCondensedRoman-Regular",
        "CabinCondensedRoman-Medium",
        "CabinCondensedRoman-SemiBold",
        "CabinCondensedRoman-Bold",
    ]
    _check_ps_instance_names(condensed_roman, desired_condensed_roman_ps_names)

    desired_condensed_italic_ps_names = [
        "CabinCondensedItalic-Regular",
        "CabinCondensedItalic-Medium",
        "CabinCondensedItalic-SemiBold",
        "CabinCondensedItalic-Bold",
    ]
    _check_ps_instance_names(condensed_italic, desired_condensed_italic_ps_names)


def test_gen_stat_update_fvar_instances_3(var_fonts3):
    gen_stat_tables(var_fonts3, axis_order=["wdth", "wght", "ital"])
    roman, italic = var_fonts3

    desired_roman_ps_names = [
        "CabinRoman-Regular",
        "CabinRoman-Medium",
        "CabinRoman-SemiBold",
        "CabinRoman-Bold",
    ]
    _check_ps_instance_names(roman, desired_roman_ps_names)

    desired_italic_ps_names = [
        "CabinItalic-Regular",
        "CabinItalic-Medium",
        "CabinItalic-SemiBold",
        "CabinItalic-Bold",
    ]
    _check_ps_instance_names(italic, desired_italic_ps_names)


def test_gen_stat_nameid_25_vf_postscript_name_1(var_font):
    gen_stat_tables([var_font], axis_order=['wdth', 'wght'])
    assert var_font['name'].getName(25, 3, 1, 0x409).toUnicode() == "Inconsolata"


def test_gen_stat_nameid_25_vf_postscript_name_2(var_fonts):
    gen_stat_tables(var_fonts, axis_order=['wght', 'ital'])
    roman, italic = var_fonts
    assert roman['name'].getName(25, 3, 1, 0x409).toUnicode() == "RalewayRoman"
    assert italic['name'].getName(25, 3, 1, 0x409).toUnicode() == "RalewayItalic"


def test_gen_stat_nameid_25_vf_postscript_name_3(var_fonts2):
    gen_stat_tables(var_fonts2, axis_order=["wdth", "wght", "ital"])
    roman, italic, condensed_roman, condensed_italic = var_fonts2
    assert roman['name'].getName(25, 3, 1, 0x409).toUnicode() == "CabinNormalRoman"
    assert italic['name'].getName(25, 3, 1, 0x409).toUnicode() == "CabinNormalItalic"
    assert condensed_roman['name'].getName(25, 3, 1, 0x409).toUnicode() == "CabinCondensedRoman"
    assert condensed_italic['name'].getName(25, 3, 1, 0x409).toUnicode() == "CabinCondensedItalic"

 
def test_gen_stat_nameid_25_vf_postscript_name_4(var_fonts3):
    gen_stat_tables(var_fonts3, axis_order=["wdth", "wght", "ital"])
    roman, italic = var_fonts3
    assert roman['name'].getName(25, 3, 1, 0x409).toUnicode() == "CabinRoman"
    assert italic['name'].getName(25, 3, 1, 0x409).toUnicode() == "CabinItalic"


def test_gen_stat_tables_from_config(var_fonts):
    config_text = """
      Raleway[wght].ttf:
      - name: Weight
        tag: wght
        values:
        - name: Regular
          value: 400
          flags: 2
        - name: Bold
          value: 700
        - name: SemiBold
          value: 600
      - name: Italic
        tag: ital
        values:
        - name: Roman
          value: 0
          linkedValue: 1
          flags: 2

      Raleway-Italic[wght].ttf:
      - name: Weight
        tag: wght
        values:
        - name: Regular
          value: 400
          flags: 2
        - name: Bold
          value: 700
        - name: SemiBold
          value: 600
      - name: Italic
        tag: ital
        values:
        - name: Italic
          value: 1
    """
    config = yaml.load(config_text, Loader=yaml.SafeLoader)
    gen_stat_tables_from_config(config, var_fonts)
    roman, italic = var_fonts

    roman_axis_val = _get_axis_value(roman, "ital", "Roman", 0.0)
    assert roman_axis_val != None
    assert roman_axis_val.LinkedValue == 1.0

    italic_axis_val = _get_axis_value(italic, "ital", "Italic", 1.0)
    assert italic_axis_val != None

    roman_reg_val = _get_axis_value(roman, "wght", "Regular", 400)
    assert roman_reg_val != None
    assert roman_reg_val.Flags == 0x2

    ital_reg_val = _get_axis_value(italic, "wght", "Regular", 400)
    assert roman_reg_val != None
    assert roman_reg_val.Flags == 0x2

    assert _get_axis_value(roman, "wght", "Light", 300) == None
    assert _get_axis_value(italic, "wght", "Light", 300) == None

    # Check ps names
    assert roman['name'].getName(25, 3, 1, 0x409).toUnicode() == "RalewayRoman"
    assert italic['name'].getName(25, 3, 1, 0x409).toUnicode() == "RalewayItalic"
    desired_roman_ps_names = [
        "RalewayRoman-Thin",
        "RalewayRoman-ExtraLight",
        "RalewayRoman-Light",
        "RalewayRoman-Regular",
        "RalewayRoman-Medium",
        "RalewayRoman-SemiBold",
        "RalewayRoman-Bold",
        "RalewayRoman-ExtraBold",
        "RalewayRoman-Black",
    ]
    _check_ps_instance_names(roman, desired_roman_ps_names)

    desired_italic_ps_names = [
        "RalewayItalic-Thin",
        "RalewayItalic-ExtraLight",
        "RalewayItalic-Light",
        "RalewayItalic-Regular",
        "RalewayItalic-Medium",
        "RalewayItalic-SemiBold",
        "RalewayItalic-Bold",
        "RalewayItalic-ExtraBold",
        "RalewayItalic-Black",
    ]
    _check_ps_instance_names(italic, desired_italic_ps_names)



#!/usr/bin/env python3
"""
Set the glyph Order for a collection of ufos

Usage:

# Set glyph orders based on first ufo
gftools ufo-set-order font1.ufo font2.ufo font3.ufo

# Set glyph orders based on font1.ufo glyph order
gftools ufo-set-order font2.ufo font3.ufo --origin font1.ufo
"""
# TODO: we can probably order components and anchors as well!
# add these when needed
from defcon import Font
import argparse
import os


def set_glyph_order(origin, fonts):
    glyph_order = origin.glyphOrder
    for font in fonts:
        if font.glyphOrder != glyph_order:
            print(f"Updating {os.path.basename(font.path)} since glyph order is different")
        font.glyphOrder = glyph_order


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("fonts", nargs="+")
    parser.add_argument("--origin", help="Source font to set glyph order for other fonts")
    args = parser.parse_args()

    if len(args.fonts) <= 1:
        print("Single font. No need to set order")
        return

    if not args.origin:
        origin = args.fonts[0]
        fonts = args.fonts[1:]
    else:
        origin = args.origin
    
    fonts = [Font(fp) for fp in args.fonts]
    origin = Font(origin)

    set_glyph_order(origin, fonts)

    for font in fonts:
        font.save()


if __name__ == "__main__":
    main()


from gftools import fonts_public_pb2
import enum
import re

class Udhr():

  def __init__(self, key, iso639_3, iso15924, bcp47, direction, ohchr, stage, loc, name):
    self.key = key
    self.iso639_3 = iso639_3
    self.iso15924 = iso15924
    self.bcp47 = bcp47
    self.direction = direction
    self.ohchr = ohchr
    self.stage = stage
    self.loc = loc
    self.name = name

    self.title = None
    self.preamble = None
    self.articles = []

  def Parse(self, translation_data):
    if translation_data is None or self.stage < 2:
      return

    if translation_data.find('./{*}title') is not None:
      self.title = translation_data.find('./{*}title').text

    preamble_data = translation_data.find('./{*}preamble')
    if preamble_data is not None:
      if preamble_data.find('./{*}title') is not None:
        self.preamble = {
            'title':
                preamble_data.find('./{*}title').text,
            'content': [
                para.text for para in preamble_data.findall('./{*}para')
                    ],
        }

    articles_data = translation_data.findall('./{*}article')
    for article_data in articles_data:
      title_data = article_data.find('./{*}title')
      article = {
          'id':
              int(article_data.get('number')),
          'title': None if title_data is None else title_data.text,
          'content': [
              para.text for para in article_data.findall('./{*}para')
                  ],
      }
      self.articles.append(article)

  def LoadArticleOne(self, article_one):
    self.articles.append({'id': 0, 'title': None, 'content': [article_one]})

  def GetSampleTexts(self):
    extractor = SampleTextExtractor(self)
    return extractor.GetSampleTexts()


class SampleTextExtractor():

  class TextType(enum.Enum):
    GLYPHS = 1
    WORD = 2
    PHRASE = 3
    SENTENCE = 4
    PARAGRAPH = 5
    PASSAGE = 6

  def __init__(self, udhr):
    self._udhr = udhr
    self._glyphs = iter(self._GetGlyphs())
    self._words = iter(self._GetWords())
    self._paragraphs = iter(self._GetParagraphs())
    self._phrase_history = set()

    self._non_word_regex = re.compile(r'[^\w]+')
    self._space_regex = re.compile(r'\s+')
    self._non_space_regex = re.compile(r'[^\s]+')
    self._non_word_space_regex = re.compile(r'[^\w\s]+')
    self._any_regex = re.compile(r'.')

  def _DisplayLength(self, s):
    """Returns length of given string. Omits combining characters.

    Some entire scripts will not be counted; in those cases, the raw length of
    the string is returned.
    """
    word_space_length = len(self._non_word_space_regex.sub('', s))
    space_length = len(self._non_space_regex.sub('', s))
    if word_space_length == space_length:
      return len(s)
    return word_space_length

  def _GetGlyphs(self):
    seen = set()
    for article in self._udhr.articles:
      for para in article['content']:
        for ch in (self._non_word_regex.sub('', para) or self._space_regex.sub('', para)):
          ch = ch.lower()
          if ch not in seen:
            seen.add(ch)
            yield ch

  def _GetWords(self):
    if self._space_regex.search(self._udhr.articles[0]['content'][0]) is not None:
      splitter = self._space_regex
    else:
      splitter = self._non_word_regex

    seen = set()
    for article in self._udhr.articles:
      for para in article['content']:
        for s in splitter.split(para):
          if s not in seen:
            seen.add(s)
            yield s

  def _GetParagraphs(self):
    if self._udhr.preamble is not None:
      for para in self._udhr.preamble['content']:
        yield para
    for article in self._udhr.articles:
      for para in article['content']:
        yield para

  def _ExtractGlyphs(self, min_chars, max_chars):
    s = ''
    for ch in self._glyphs:
      s += ch.upper()
      if len(s) >= min_chars:
        break
      if ch != ch.upper():
        s += ch
        if len(s) >= min_chars:
          break
    return s

  def _ExtractWord(self, min_chars, max_chars):
    for iterator in [self._words, self._GetWords()]:
      for w in iterator:
        if w is None:
          continue
        if min_chars <= self._DisplayLength(w) <= max_chars:
          return w
    # Fallback to using multiple words for languages with very small words
    return self._ExtractPhrase(min_chars, max_chars)

  def _ExtractPhrase(self, min_chars, max_chars):
    for iterator in [self._paragraphs, self._GetParagraphs()]:
      for para in iterator:
        if para is None:
          continue
        for regex in [self._any_regex, self._space_regex, self._non_word_regex]:
          breaks = [-1]
          for match in regex.finditer(para, min_chars):
            breaks.append(match.start())
            phrase = para[breaks[0]+1:breaks[len(breaks)-1]]
            p_size = self._DisplayLength(phrase)
            while p_size > max_chars and len(breaks) > 1:
              breaks.pop()
              phrase = para[breaks[0]+1:breaks[len(breaks)-1]]
              p_size = self._DisplayLength(phrase)
            if min_chars <= p_size and phrase not in self._phrase_history:
              self._phrase_history.add(phrase)
              return phrase
    return self._ExtractParagraph(min_chars, max_chars)

  def _ExtractSentence(self, min_chars, max_chars):
    # Sentence delimination may differ between scripts, so tokenizing on spaces
    # would be unreliable. Prefer to use _ExtractPhrase.
    return self._ExtractPhrase(min_chars, max_chars)

  def _ExtractParagraph(self, min_chars, max_chars):
    for iterator in [self._paragraphs, self._GetParagraphs()]:
      for para in iterator:
        if para is None:
          continue
        if min_chars <= self._DisplayLength(para) <= max_chars:
          return para
    # Paragraphs likely insufficient length; try combining into passages
    return self._ExtractPassage(min_chars, max_chars)

  def _ExtractPassage(self, min_chars, max_chars):
    p = []
    p_size = 0
    while p_size < min_chars:
      for iterator in [self._paragraphs, self._GetParagraphs()]:
        for para in iterator:
          if para is None:
            continue
          p.append(para)
          p_size = self._DisplayLength(' '.join(p))
          if max_chars < p_size:
            p.pop()
          elif min_chars <= p_size:
            return '\n'.join(p)
    assert len(p) > 0, 'Unable to extract passage: ' + self._udhr.key
    if len(p) == 0:
      p.append([p for p in self._GetParagraphs()][0])
    return '\n'.join(p)

  def _Get(self, text_type, **kwargs):
    if 'char_count' in kwargs:
      min_chars = kwargs['char_count']
      max_chars = kwargs['char_count']
    else:
      min_chars = kwargs['min_chars']
      max_chars = kwargs['max_chars']
    if text_type == self.TextType.GLYPHS:
      return self._ExtractGlyphs(min_chars, max_chars)
    if text_type == self.TextType.WORD:
      return self._ExtractWord(min_chars, max_chars)
    if text_type == self.TextType.PHRASE:
      return self._ExtractPhrase(min_chars, max_chars)
    if text_type == self.TextType.SENTENCE:
      return self._ExtractSentence(min_chars, max_chars)
    if text_type == self.TextType.PARAGRAPH:
      return self._ExtractParagraph(min_chars, max_chars)
    if text_type == self.TextType.PASSAGE:
      return self._ExtractPassage(min_chars, max_chars)
    raise Exception('Unsupported text type: ' + text_type)

  def GetSampleTexts(self):
    sample_text = fonts_public_pb2.SampleTextProto()
    sample_text.masthead_full = self._Get(self.TextType.GLYPHS, char_count = 4)
    sample_text.masthead_partial = self._Get(self.TextType.GLYPHS, char_count = 2)
    sample_text.styles = self._Get(self.TextType.PHRASE, min_chars = 40, max_chars = 60)
    sample_text.tester = self._Get(self.TextType.PHRASE, min_chars = 60, max_chars = 90)
    sample_text.poster_sm = self._Get(self.TextType.PHRASE, min_chars = 10, max_chars = 17)
    sample_text.poster_md = self._Get(self.TextType.PHRASE, min_chars = 6, max_chars = 12)
    sample_text.poster_lg = self._Get(self.TextType.WORD, min_chars = 3, max_chars = 8)
    sample_text.specimen_48 = self._Get(self.TextType.SENTENCE, min_chars = 50, max_chars = 80)
    sample_text.specimen_36 = self._Get(self.TextType.PARAGRAPH, min_chars = 100, max_chars = 120)
    sample_text.specimen_32 = self._Get(self.TextType.PARAGRAPH, min_chars = 140, max_chars = 180)
    sample_text.specimen_21 = self._Get(self.TextType.PASSAGE, min_chars = 300, max_chars = 500)
    sample_text.specimen_16 = self._Get(self.TextType.PASSAGE, min_chars = 550, max_chars = 750)
    return sample_text




#!/usr/bin/env python3
"""Check a font family using Google Fonts QA tools.

Examples:
Check a local family against the same family hosted on Google Fonts:
`gftools qa -f [fonts.ttf] -gfb -a -o qa`

Check a local family against another local family and generate reports
for Font Diffenator only:
`gftools qa -f [fonts_a.ttf] -fb [fonts_b.ttf] --diffenator -o qa`

Check a local family against the same family hosted on Google Fonts and
generate reports for Diffbrowsers only:
`gftools qa -f [fonts.ttf] -gf --diffbrowsers -o qa

Compare a pull request against the same family hosted on Google Fonts:
`gftools qa -pr www.github.com/user/repo/pull/1 -gfb -a -o qa`

Compare a github folder of fonts against the same family hosted on Google
Fonts:
`gftools qa -gh www.github.com/user/repo/tree/fonts/ttf -gfb -a -o qa`
"""
from fontTools.ttLib import TTFont
import argparse
import shutil
import os
import subprocess
import logging
import re
from gftools.utils import (
    download_family_from_Google_Fonts,
    download_files_in_github_pr,
    download_files_in_github_dir,
    Google_Fonts_has_family,
    mkdir,
)
from gftools.html import HtmlProof, HtmlDiff
try:
    from diffenator.diff import DiffFonts
    from diffenator.font import DFont
    from diffbrowsers.utils import load_browserstack_credentials as bstack_creds
except ModuleNotFoundError:
    raise ModuleNotFoundError(("gftools was installed without the QA "
        "dependencies. To install the dependencies, see the ReadMe, "
        "https://github.com/googlefonts/gftools#installation"))
from gftools.github import GitHubClient

__version__ = "2.1.3"
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def load_browserstack_credentials():
    """Return the user's Browserstack credentials"""
    credentials = bstack_creds()
    if not credentials:
        username = os.environ.get("BSTACK_USERNAME")
        access_key = os.environ.get("BSTACK_ACCESS_KEY")
        if all([username, access_key]):
            return (username, access_key)
        return False
    return credentials


class FontQA:

    def __init__(self, fonts, fonts_before=None, out="out"):
        self.fonts = fonts
        self.fonts_before = fonts_before

        self.instances = self._instances_in_fonts(self.fonts)
        self.instances_before = self._instances_in_fonts(self.fonts_before)
        self.matching_instances = self._matching_instances()

        self._bstack_auth = load_browserstack_credentials()
        self.out = out

    def _instances_in_fonts(self, ttfonts):
        """Get all font instances from a collection of fonts.

        This function works for both a static and variable font collections.
        If a font is variable, it will retrieve the font's instances
        using the fvar table. If a font is static, it will only return a
        single instance by using the font's filename.
        """
        if not ttfonts:
            return None
        results = {}
        for ttfont in ttfonts:
            if "fvar" in ttfont:
                for instance in ttfont['fvar'].instances:
                    nameid = instance.subfamilyNameID
                    name = ttfont['name'].getName(nameid, 3, 1, 1033).toUnicode()
                    name = name.replace(" ", "")
                    results[name] = {
                        "coordinates": instance.coordinates,
                        "filename": ttfont.reader.file.name
                    }
            else:
                filename = os.path.basename(ttfont.reader.file.name)
                name = filename.split("-")[1]
                name = re.sub(".ttf|.otf", "", name)
                results[name] = {
                    "coordinates": {"wght": ttfont['OS/2'].usWeightClass},
                    "filename": ttfont.reader.file.name
                }
        return results

    def _matching_instances(self):
        if not self.fonts_before:
            logger.info(
                "No regression checks possible since there are no previous fonts."
            )
            return None
        shared = set(self.instances_before.keys()) & set(self.instances.keys())
        new = set(self.instances.keys()) - set(self.instances_before.keys())
        missing = set(self.instances_before.keys()) - set(self.instances.keys())
        if new:
            logger.warning("New fonts: {}".format(", ".join(new)))
        if missing:
            logger.warning("Missing fonts: {}".format(", ".join(missing)))
        if not shared:
            raise Exception(
                (
                    "Cannot find matching fonts!\n"
                    "fonts: [{}]\nfonts_before: [{}]".format(
                        ", ".join(set(self.instances.keys())),
                        ", ".join(set(self.instances_before.keys()))
                     )
                )
            )
        return shared

    def diffenator(self, **kwargs):
        logger.info("Running Diffenator")
        dst = os.path.join(self.out, "Diffenator")
        mkdir(dst)
        for style in self.matching_instances:
            font_before = DFont(self.instances_before[style]['filename'])
            font_after = DFont(self.instances[style]['filename'])
            out = os.path.join(dst, style)
            if font_after.is_variable and not font_before.is_variable:
                font_after.set_variations_from_static(font_before)

            elif not font_after.is_variable and font_before.is_variable:
                font_before.set_variations_from_static(font_after)

            elif font_after.is_variable and font_before.is_variable:
                coordinates = self.instances_before[style]['coordinates']
                font_after.set_variations(coordinates)
                font_before.set_variations(coordinates)

            # TODO add settings
            diff = DiffFonts(font_before, font_after, {"render_diffs": True})
            diff.to_gifs(dst=out)
            diff.to_txt(20, os.path.join(out, "report.txt"))
            diff.to_md(20, os.path.join(out, "report.md"))
            diff.to_html(20, os.path.join(out, "report.html"), image_dir=".")

    def diffbrowsers(self, **kwargs):
        """Test fonts on GFR regression and take screenshots using
        diffbrowsers. A browserstack account is required."""
        logger.info("Running Diffbrowsers")
        if not self._bstack_auth:
            logger.info("Skipping. No Browserstack credentials. "
                    "See https://github.com/googlefonts/"
                    "diffbrowsers#installation on how to add them.")
            return
        dst = os.path.join(self.out, "Diffbrowsers")
        mkdir(dst)

        html = HtmlDiff(
            out=dst,
            fonts_before=[f.reader.file.name for f in self.fonts_before],
            fonts_after=[f.reader.file.name for f in self.fonts],
        )
        html.build_pages(["waterfall.html", "text.html"])
        html.build_pages(["glyphs.html"], pt_size=16)
        html.save_imgs()

    def fontbakery(self):
        logger.info("Running Fontbakery")
        out = os.path.join(self.out, "Fontbakery")
        mkdir(out)
        cmd = (
            ["fontbakery", "check-googlefonts", "-l", "INFO", "--succinct"]
            + [f.reader.file.name for f in self.fonts]
            + ["-C"]
            + ["--ghmarkdown", os.path.join(out, "report.md")]
        )
        subprocess.call(cmd)

    def plot_glyphs(self):
        logger.info("Running plot glyphs")
        out = os.path.join(self.out, "plot_glyphs")
        mkdir(out)
        fonts = [f.reader.file.name for f in self.fonts]
        for font in fonts:
            font_filename = os.path.basename(font)[:-4]
            dfont = DFont(font)
            if dfont.is_variable:
                for _, coords in dfont.instances_coordinates.items():
                    dfont.set_variations(coords)
                    img_out = os.path.join(
                        out,
                        "%s_%s.png"
                        % (font_filename, self._instance_coords_to_filename(coords)),
                    )
                    dfont.glyphs.to_png(img_out, limit=100000)
            else:
                img_out = os.path.join(out, font_filename + ".png")
                dfont.glyphs.to_png(dst=img_out)

    def _instance_coords_to_filename(self, d):
        name = ""
        for k, v in d.items():
            name += "{}_{}_".format(k, v)
        return name[:-1]

    def browser_previews(self, **kwargs):
        """Use GFR and diffbrowsers to take screenshots of how the fonts
        will look on different browsers. A Browserstack account is
        required."""
        logger.info("Running browser previews")
        if not self._bstack_auth:
            logger.info("Skipping. No Browserstack credentials. "
                    "See https://github.com/googlefonts/"
                    "diffbrowsers#installation on how to add them.")
            return
        out = os.path.join(self.out, "browser_previews")
        mkdir(out)
        html = HtmlProof(
            out=out,
            fonts=[f.reader.file.name for f in self.fonts]
        )
        html.build_pages(["waterfall.html", "text.html"])
        html.build_pages(["glyphs.html"], pt_size=16)
        html.save_imgs()

    def googlefonts_upgrade(self):
        self.fontbakery()
        self.diffenator()
        self.diffbrowsers()

    def googlefonts_new(self):
        self.fontbakery()
        self.plot_glyphs()
        self.browser_previews()

    def post_to_github(self, url):
        """Post Fontbakery report as a new issue or as a comment to an open
        PR"""
        # Parse url tokens
        url_split = url.split("/")
        repo_owner = url_split[3]
        repo_name = url_split[4]
        issue_number = url_split[-1] if "pull" in url else None

        fontbakery_report = os.path.join(self.out, "Fontbakery", "report.md")
        if not os.path.isfile(fontbakery_report):
            logger.warning(
                "Cannot Post Github message because no Fontbakery report exists"
            )
            return
        
        client = GitHubClient(repo_owner, repo_name)

        with open(fontbakery_report) as doc:
            msg = doc.read()
            if issue_number:
                client.create_issue_comment(issue_number, msg)
            else:
                client.create_issue("Google Font QA report", msg)


def family_name_from_fonts(fonts):
    results = []
    for font in fonts:
        family_name = font["name"].getName(1, 3, 1, 1033)
        typo_family_name = font["name"].getName(16, 3, 1, 1033)

        if typo_family_name:
            results.append(typo_family_name.toUnicode())
        elif family_name:
            results.append(family_name.toUnicode())
        else:
            raise Exception(
                "Font: {} has no family name records".format(
                    os.path.basename(font.reader.file.name)
                )
            )
    if len(set(results)) > 1:
        raise Exception("Multiple family names found: [{}]".format(", ".join(results)))
    return results[0]


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    font_group = parser.add_argument_group(title="Fonts to qa")
    font_input_group = font_group.add_mutually_exclusive_group(required=True)
    font_input_group.add_argument("-f", "--fonts", nargs="+",
        help="Paths to fonts")
    font_input_group.add_argument("-pr", "--pull-request",
        help="Get fonts from a Github pull request")
    font_input_group.add_argument("-gh", "--github-dir",
        help="Get fonts from a Github directory")
    font_input_group.add_argument("-gf", "--googlefonts",
        help="Get fonts from Google Fonts")

    font_before_group = parser.add_argument_group(title="Fonts before input")
    font_before_input_group = font_before_group.add_mutually_exclusive_group(
        required=False
    )
    font_before_input_group.add_argument(
        "-fb", "--fonts-before", nargs="+",
        help="Paths to previous fonts"
    )
    font_before_input_group.add_argument("-prb", "--pull-request-before",
        help="Get previous fonts from a Github pull request")
    font_before_input_group.add_argument("-ghb", "--github-dir-before",
        help="Get previous fonts from a Github dir")
    font_before_input_group.add_argument(
        "-gfb",
        "--googlefonts-before",
        action="store_true",
        help="Get previous fonts from Google Fonts",
    )

    check_group = parser.add_argument_group(title="QA checks")
    check_group.add_argument(
        "-a",
        "--auto-qa",
        action="store_true",
        help="Check fonts against against the same fonts hosted on Google Fonts",
    )
    check_group.add_argument(
        "--diffenator", action="store_true", help="Run Fontdiffenator"
    )
    check_group.add_argument(
        "--diffbrowsers", action="store_true", help="Run Diffbrowsers"
    )
    check_group.add_argument(
        "--fontbakery", action="store_true", help="Run FontBakery"
    )
    check_group.add_argument(
        "--plot-glyphs",
        action="store_true",
        help="Gen images of full charset, useful for new familes",
    )
    check_group.add_argument(
        "--browser-previews",
        action="store_true",
        help="Gen images on diff browsers, useful for new families",
    )
    check_group.add_argument(
        "-dm", "--diff-mode", choices=("weak", "normal", "strict"), default="normal"
    )
    parser.add_argument("-re", "--filter-fonts", help="Filter fonts by regex")
    parser.add_argument(
        "-o", "--out", default="out", help="Output path for check results"
    )
    parser.add_argument(
        "-ogh",
        "--out-github",
        action="store_true",
        help=(
            "Post report data to either the pull request as a comment "
            "open a new issue. This can only be used if fonts have been "
            "fetched from either a pull request or github dir."
        ),
    )
    parser.add_argument(
        "--out-url",
        help=(
            "Post report data to a github pr. This can be used with any font "
            "fetching method."
        )
    )
    parser.add_argument("--version", action="version", version=__version__)
    args = parser.parse_args()
    if args.out_github and not any([args.pull_request, args.github_dir]):
        raise Exception(
            "Cannot upload results to a github issue or pr. "
            "Font input must either a github dir or a pull request"
        )
    if not any([args.auto_qa,
                args.fontbakery,
                args.plot_glyphs,
                args.diffbrowsers,
                args.diffenator,
                args.browser_previews]):
        raise Exception("Terminating. No checks selected. Run gftools qa "
                        "--help to see all possible commands.")


    # Retrieve fonts and store in out dir
    mkdir(args.out)
    fonts_dir = os.path.join(args.out, "fonts")
    mkdir(fonts_dir)
    if args.fonts:
        [shutil.copy(f, fonts_dir) for f in args.fonts]
        fonts = args.fonts
    elif args.pull_request:
        fonts = download_files_in_github_pr(
            args.pull_request,
            fonts_dir,
            ignore_static_dir=False,
        )
        if not fonts:
            logger.info("No fonts found in pull request. Skipping")
            return
    elif args.github_dir:
        fonts = download_files_in_github_dir(args.github_dir, fonts_dir)
        if not fonts:
            logger.info("No fonts found in github dir. Skipping")
            return
    elif args.googlefonts:
        fonts = download_family_from_Google_Fonts(args.googlefonts, fonts_dir)

    if args.filter_fonts:
        re_filter = re.compile(args.filter_fonts)
        fonts = [f for f in fonts if re_filter.search(f)]

    ttfonts = [TTFont(f) for f in fonts if f.endswith((".ttf", ".otf"))
               and "static" not in f]
    family_name = family_name_from_fonts(ttfonts)
    family_on_gf = Google_Fonts_has_family(family_name)

    # Retrieve fonts_before and store in out dir
    fonts_before = None
    if any([args.fonts_before, args.pull_request_before, args.github_dir_before]) or \
           (args.googlefonts_before and family_on_gf):
        fonts_before_dir = os.path.join(args.out, "fonts_before")
        mkdir(fonts_before_dir, overwrite=False)
    if args.fonts_before:
        [shutil.copy(f, fonts_before_dir) for f in args.fonts_before]
        fonts_before = args.fonts_before
    elif args.pull_request_before:
        fonts_before = download_files_in_github_pr(
            args.pull_request_before,
            fonts_before_dir,
            ignore_static_dir=False
        )
    elif args.github_dir_before:
        fonts_before = download_files_in_github_dir(
            args.github_dir_before, fonts_before_dir
        )
    elif args.googlefonts_before and family_on_gf:
        fonts_before = download_family_from_Google_Fonts(
            family_name, fonts_before_dir
        )

    if fonts_before:
        ttfonts_before = [TTFont(f) for f in fonts_before if f.endswith((".ttf", ".otf"))
                          and "static" not in f]
        qa = FontQA(ttfonts, ttfonts_before, args.out)
    else:
        qa = FontQA(ttfonts, out=args.out)

    if args.auto_qa and family_on_gf:
        qa.googlefonts_upgrade()
    elif args.auto_qa and not family_on_gf:
        qa.googlefonts_new()
    if args.plot_glyphs:
        qa.plot_glyphs()
    if args.browser_previews:
        qa.browser_previews()
    if args.fontbakery:
        qa.fontbakery()
    if args.diffenator:
        qa.diffenator()
    if args.diffbrowsers:
        qa.diffbrowsers()

    if args.out_url:
        qa.post_to_github(args.out_url)
    elif args.out_github and args.pull_request:
        qa.post_to_github(args.pull_request)
    elif args.out_github and args.github_dir:
        qa.post_to_github(args.github_dir)


if __name__ == "__main__":
    main()
